from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING, ParamSpec, TypeVar

from brikk.returns import Error, Nothing, Ok, Option, Result, Some

if TYPE_CHECKING:
    from collections.abc import Callable

P = ParamSpec("P")
T = TypeVar("T")
E = TypeVar("E", bound=Exception)


def result_of(
    func: Callable[P, T], *args: P.args, **kwargs: P.kwargs
) -> Result[T, Exception]:
    """Execute a function and wrap its result in a Result.

    :param Callable[P, T] func: The function to execute.
    :param args: Positional arguments for the function.
    :param kwargs: Keyword arguments for the function.

    :Example:

    >>> def div(a, b):
    ...     return a / b
    >>> result_of(div, 4, 2)
    Ok(2.0)
    >>> result_of(div, 4, 0)
    Error(ZeroDivisionError(...))

    :return: Ok(value) if successful, Error(exception) otherwise.
    :rtype: Result[T, Exception]
    """
    return as_result(func, (Exception,))(*args, **kwargs)


def optional_of(
    func: Callable[P, T | None], *args: P.args, **kwargs: P.kwargs
) -> Option[T]:
    """Execute a function and wrap its result in an Option.

    :param Callable[P, T | None] func: The function to execute.
    :param args: Positional arguments for the function.
    :param kwargs: Keyword arguments for the function.

    :Example:

    >>> def find(x):
    ...    return x if x > 0 else None
    >>> optional_of(find, 5)
    Some(5)
    >>> optional_of(find, -1)
    Nothing()

    :return: Some(value) if not None, Nothing otherwise.
    :rtype: Option[T]
    """
    return as_optional(func)(*args, **kwargs)


def as_result(
    func: Callable[P, T],
    exceptions: tuple[type[E], ...] = (Exception,),
) -> Callable[P, Result[T, E]]:
    """Wrap a function so its result is returned as a Result.

    :param Callable[P, T] func: The function to wrap.
    :param tuple exceptions: Exception types to catch.

    :Example:

    >>> def div(a, b):
    ...     return a / b
    >>> as_result(div)(4, 2)
    Ok(2.0)
    >>> as_result(div)(4, 2)
    Error(ZeroDivisionError(...))

    :return: Wrapped function returning Ok or Error.
    :rtype: Callable[P, Result[T, E]]
    """

    @wraps(func)
    def _wrapped(*args: P.args, **kwargs: P.kwargs) -> Result[T, E]:
        try:
            return Ok(func(*args, **kwargs))
        except exceptions as e:
            return Error(e)

    return _wrapped


def as_optional(func: Callable[P, T | None]) -> Callable[P, Option[T]]:
    """Wrap a function so its result is returned as an Option.

    :param Callable[P, T | None] func: The function to wrap.

    :Example:

    >>> @as_optional
    ... def find(x):
    ...     return x if x > 0 else None
    >>> find(5)
    Some(5)
    >>> find(-1)
    Nothing()

    :return: Wrapped function returning Some or Nothing.
    :rtype: Callable[P, Option[T]]
    """

    @wraps(func)
    def _wrapped(*args: P.args, **kwargs: P.kwargs) -> Option[T]:
        result = func(*args, **kwargs)
        if result is None:
            return Nothing()
        return Some(result)

    return _wrapped


def safe(
    exceptions: tuple[type[E], ...] = (Exception,),
) -> Callable[[Callable[P, T]], Callable[P, Result[T, E]]]:
    """Return a decorator that wraps a function to return a Result.

    :param tuple exceptions: Exception types to catch.

    :Example:

    >>> @safe((ZeroDivisionError,))
    ... def div(a, b):
    ...     return a / b
    >>> div(4, 2)
    Ok(2.0)
    >>> div(4, 0)
    Error(ZeroDivisionError(...))

    :return: Decorator for wrapping functions to return Result.
    :rtype: Callable[[Callable[P, T]], Callable[P, Result[T, E]]]
    """

    def _make(func: Callable[P, T]):
        return as_result(func, exceptions)

    return _make



class ReturnsError(Exception): ...


class OptionError(ReturnsError): ...


class OptionUnwrapError(OptionError): ...


class OptionExpectError(OptionError): ...


class ResultError(ReturnsError): ...


class ResultUnwrapError(ResultError): ...


class ResultExpectError(ResultError): ...



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    TypeAlias,
    TypeGuard,
    TypeVar,
    cast,
)

from brikk.returns._errors import OptionExpectError, OptionUnwrapError

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator

    from brikk.returns import Error, Ok, Result


T = TypeVar("T")
E = TypeVar("E")
U = TypeVar("U")

Option: TypeAlias = "Some[T] | Nothing[T]"


def is_some(result: Option[T]) -> TypeGuard[Some[T]]:
    return result.is_some()


def is_none(result: Option[T]) -> TypeGuard[Nothing[T]]:
    return result.is_none()


class _OptionBase(Generic[T]):
    def __init__(self, **kwargs) -> None:
        if "value" in kwargs:
            self.value = cast(T, kwargs["value"])

    def is_some(self) -> bool:
        return hasattr(self, "value")

    def is_some_and(self, func: Callable[[T], bool]) -> bool:
        return self.is_some() and func(self.value)

    def is_none(self) -> bool:
        return not hasattr(self, "value")

    def is_none_or(self, func: Callable[[T], bool]) -> bool:
        return self.is_none() or func(self.value)

    def expect(self, msg: str) -> T:
        if self.is_some():
            return self.value
        raise OptionExpectError(msg)

    def unwrap(self) -> T:
        if self.is_some():
            return self.value
        raise OptionUnwrapError

    def unwrap_or(self, default: T) -> T:
        if self.is_some():
            return self.value
        return default

    def unwrap_or_else(self, func: Callable[[], T]) -> T:
        if self.is_some():
            return self.value
        return func()

    def map(self, func: Callable[[T], U]) -> Option[U]:
        if self.is_some():
            return Some(func(self.value))
        return Nothing()

    def inspect(self, func: Callable[[T], None]) -> Option[T]:
        if self.is_some():
            func(self.value)
        return Nothing()

    def map_or(self, default: U, func: Callable[[T], U]) -> U:
        if self.is_some():
            return func(self.value)
        return default

    def map_or_else(self, default: Callable[[], U], func: Callable[[T], U]) -> U:
        if self.is_some():
            return func(self.value)
        return default()

    def ok_or(self, error: E) -> Result[T, E]:
        from brikk.returns import Error, Ok

        if self.is_some():
            return Ok(self.value)
        return Error(error)

    def ok_or_else(self, error: Callable[[], E]) -> Result[T, E]:
        from brikk.returns import Error, Ok

        if self.is_some():
            return Ok(self.value)
        return Error(error())

    def iter(self) -> Iterator[T]:
        if self.is_some():
            yield self.value

    def and_(self, option: Option[U]) -> Option[U]:
        if self.is_some():
            return option
        return Nothing()

    def and_then(self, func: Callable[[T], Option[U]]) -> Option[U]:
        if self.is_some():
            return func(self.value)
        return Nothing()

    def filter(self, predicate: Callable[[T], bool]) -> Option[T]:
        if self.is_some_and(predicate):
            return Some(self.value)
        return Nothing()

    def or_(self, default: Option[T]) -> Option[T]:
        if self.is_none():
            return default
        return Some(self.value)

    def or_else(self, default: Callable[[], Option[T]]) -> Option[T]:
        if self.is_some():
            return Some(self.value)
        return default()

    def xor(self, option: Option[T]) -> Option[T]:
        if self.is_some() != option.is_some():
            return Some(self.value) if self.is_some() else Some(option.value)
        return Nothing()


class Some(_OptionBase[T]):
    __match_args__ = ("value",)
    __slots__ = ("value",)

    value: T

    def __init__(self, value: T) -> None:
        super().__init__(value=value)

    def __repr__(self) -> str:
        return f"Some({self.value!r})"

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Some):
            return other.value == self.value
        return False

    def flatten(self: Some[Option[U]] | Some[Some[U]] | Some[Nothing[U]]) -> Option[U]:
        return self.and_then(lambda val: val)

    def transpose(
        self: Some[Result[U, E]] | Some[Ok[U, E]] | Some[Error[U, E]],
    ) -> Result[Option[U], E]:
        from brikk.returns import Error, Ok

        match self.value:
            case Ok(val):
                return Ok(Some(val))
            case Error(err):
                return Error(err)


class Nothing(_OptionBase[T]):
    __slots__ = tuple()

    def __init__(self) -> None:
        super().__init__()

    def __repr__(self) -> str:
        return "Nothing()"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Nothing)

    def flatten(
        self: Nothing[Option[U]] | Nothing[Some[U]] | Nothing[Nothing[U]],
    ) -> Option[U]:
        return self.and_then(lambda val: val)

    def transpose(
        self: Nothing[Result[U, E]] | Nothing[Ok[U, E]] | Nothing[Error[U, E]],
    ) -> Result[Option[U], E]:
        from brikk.returns import Ok

        return Ok(Nothing())



from brikk.returns._errors import (
    OptionError,
    OptionExpectError,
    OptionUnwrapError,
    ResultError,
    ResultExpectError,
    ResultUnwrapError,
    ReturnsError,
)
from brikk.returns._option import (
    Nothing,
    Option,
    Some,
    is_none,
    is_some,
)
from brikk.returns._result import (
    Error,
    Ok,
    Result,
    is_err,
    is_ok,
)
from brikk.returns._utils import (
    as_optional,
    as_result,
    optional_of,
    result_of,
    safe,
)

__all__ = [
    "Error",
    "Nothing",
    "Ok",
    "Option",
    "OptionError",
    "OptionExpectError",
    "OptionUnwrapError",
    "Result",
    "ResultError",
    "ResultExpectError",
    "ResultUnwrapError",
    "ReturnsError",
    "Some",
    "as_optional",
    "as_result",
    "is_err",
    "is_none",
    "is_ok",
    "is_some",
    "optional_of",
    "result_of",
    "safe",
]



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    TypeAlias,
    TypeGuard,
    TypeVar,
    cast,
)

from brikk.returns._errors import ResultExpectError, ResultUnwrapError

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator

    from brikk.returns import Nothing, Option, Some


T = TypeVar("T")
E = TypeVar("E")
U = TypeVar("U")
F = TypeVar("F")
_AnyException = TypeVar("_AnyException", bound=Exception)


def is_ok(result: Result[T, E]) -> TypeGuard[Ok[T, E]]:
    return result.is_ok()


def is_err(result: Result[T, E]) -> TypeGuard[Error[T, E]]:
    return result.is_err()


Result: TypeAlias = "Ok[T, E] | Error[T, E]"


class _ResultBase(Generic[T, E]):
    def __init__(self, **kwargs) -> None:
        if "value" in kwargs:
            self.value = cast(T, kwargs["value"])
        if "error" in kwargs:
            self.error = cast(E, kwargs["error"])

    def is_ok(self) -> bool:
        return hasattr(self, "value")

    def is_ok_and(self, func: Callable[[T], bool]) -> bool:
        return self.is_ok() and func(self.value)

    def is_err(self) -> bool:
        return hasattr(self, "error")

    def is_err_and(self, func: Callable[[E], bool]) -> bool:
        return self.is_err() and func(self.error)

    def ok(self) -> Option[T]:
        from brikk.returns import Nothing, Some

        if hasattr(self, "value"):
            return Some(self.value)
        return Nothing()

    def err(self) -> Option[E]:
        from brikk.returns import Nothing, Some

        if hasattr(self, "error"):
            return Some(self.error)
        return Nothing()

    def map(self, func: Callable[[T], U]) -> Result[U, E]:
        if self.is_ok():
            return Ok(func(self.value))
        return Error(self.error)

    def map_or(self, default: U, func: Callable[[T], U]) -> U:
        if self.is_ok():
            return func(self.value)
        return default

    def map_or_else(self, default: Callable[[E], U], func: Callable[[T], U]) -> U:
        if self.is_ok():
            return func(self.value)
        return default(self.error)

    def map_err(self, func: Callable[[E], F]) -> Result[T, F]:
        if self.is_err():
            return Error(func(self.error))
        return Ok(self.value)

    def inspect(self, func: Callable[[T], None]) -> Result[T, E]:
        if self.is_ok():
            func(self.value)
        return self  # type: ignore

    def inspect_err(self, func: Callable[[E], None]) -> Result[T, E]:
        if self.is_err():
            func(self.error)
        return self  # type: ignore

    def iter(self) -> Iterator[T]:
        if self.is_ok():
            yield self.value

    def expect(self, msg: str) -> T:
        if self.is_ok():
            return self.value

        _msg = f"{msg}: {self.error}"
        raise ResultExpectError(_msg)

    def unwrap(self) -> T:
        if self.is_ok():
            return self.value

        _msg = f"{self.error}"
        raise ResultUnwrapError(_msg)

    def expect_err(self, msg: str) -> E:
        if self.is_err():
            return self.error

        _msg = f"{msg}: {self.value}"
        raise ResultExpectError(_msg)

    def unwrap_err(self) -> E:
        if self.is_err():
            return self.error

        _msg = f"{self.value}"
        raise ResultUnwrapError(_msg)

    def and_(self, result: Result[U, E]) -> Result[U, E]:
        if self.is_ok():
            return result
        return Error(self.error)

    def and_then(self, func: Callable[[T], Result[U, E]]) -> Result[U, E]:
        if self.is_ok():
            return func(self.value)
        return Error(self.error)

    def or_(self, default: Result[T, F]) -> Result[T, F]:
        if self.is_err():
            return default
        return Ok(self.value)

    def or_else(self, default: Callable[[E], Result[T, F]]) -> Result[T, F]:
        if self.is_err():
            return default(self.error)
        return Ok(self.value)

    def unwrap_or(self, default: T) -> T:
        if self.is_ok():
            return self.value
        return default

    def unwrap_or_else(self, default: Callable[[E], T]) -> T:
        if self.is_ok():
            return self.value
        return default(self.error)

    def unwrap_or_raise(self: _ResultBase[T, _AnyException]) -> T:
        if self.is_err():
            raise self.error
        return self.value


class Ok(_ResultBase[T, E]):
    __match_args__ = ("value",)
    __slots__ = ("value",)

    value: T

    def __init__(self, value: T) -> None:
        super().__init__(value=value)

    def __repr__(self) -> str:
        return f"Ok({self.value!r})"

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Ok):
            return other.value == self.value
        return False

    def flatten(
        self: Ok[Ok[U, E], E] | Ok[Error[U, E], E] | Ok[Result[U, E], E],
    ) -> Result[U, E]:
        return self.and_then(lambda val: val)

    def transpose(
        self: Ok[Some[U], E] | Ok[Nothing[U], E] | Ok[Option[U], E],
    ) -> Option[Result[U, E]]:
        from brikk.returns import Nothing, Some, is_some

        if is_some(self.value):
            return Some(Ok(self.value.value))
        return Nothing()


class Error(_ResultBase[T, E]):
    __match_args__ = ("error",)
    __slots__ = ("error",)

    error: E

    def __init__(self, error: E) -> None:
        super().__init__(error=error)

    def __repr__(self) -> str:
        return f"Error({self.error!r})"

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Error):
            return other.error == self.error
        return False

    def flatten(
        self: Error[Ok[U, E], E] | Error[Error[U, E], E] | Error[Result[U, E], E],
    ) -> Result[U, E]:
        return self.and_then(lambda val: val)

    def transpose(
        self: Error[Some[U], E] | Error[Nothing[U], E] | Error[Option[U], E],
    ) -> Option[Result[U, E]]:
        from brikk.returns import Some

        return Some(Error(self.error))
